﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows;

namespace AudioTools
{
    static public class WaveFormat
    {
        //---------------------------------------------------------------------
        //  WAVEフォーマット情報
        //---------------------------------------------------------------------
        public class WaveFormatInfo
        {
            public string Filename = string.Empty;
            public bool IsValid = false;
            public short FormatID = 0;
            public short ChannelCount = 0;
            public int SampleRate = 0;
            public int BytesPerSecond = 0;
            public short BlockAlign = 0;
            public short Bit = 0;
            public double Duration = 0;
            public int SampleCount = 0;
            public int SampleDataSize = 0;
            public bool HasLoop = false;
            public int LoopStartSample = 0;
            public int LoopEndSample = 0;
            public long SampleChunkPos = 0;

            //-----------------------------------------------------------------
            //  コンストラクタ
            //-----------------------------------------------------------------
            public WaveFormatInfo()
            {
            }

            //-----------------------------------------------------------------
            //  コンストラクタ
            //-----------------------------------------------------------------
            public WaveFormatInfo(WaveFormatInfo info)
            {
                Filename = info.Filename;
                IsValid = info.IsValid;
                FormatID = info.FormatID;
                ChannelCount = info.ChannelCount;
                SampleRate = info.SampleRate;
                BytesPerSecond = info.BytesPerSecond;
                BlockAlign = info.BlockAlign;
                Bit = info.Bit;
                Duration = info.Duration;
                SampleCount = info.SampleCount;
                SampleDataSize = info.SampleDataSize;
                HasLoop = info.HasLoop;
                LoopStartSample = info.LoopStartSample;
                LoopEndSample = info.LoopEndSample;
                SampleChunkPos = info.SampleChunkPos;
            }
        }

        //---------------------------------------------------------------------
        //  フォーマット情報の取得
        //---------------------------------------------------------------------
        public static WaveFormatInfo GetFormatInfo(string filename)
        {
            //  フォーマット情報の作成
            WaveFormatInfo info = new WaveFormatInfo();
            info.Filename = filename;
            info.IsValid = false;

            //  ファイルを読み込んで解析
            using (FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read))
            {
                try
                {
                    long fmtChunkPos = 0;
                    long smplChunkPos = 0;
                    byte[] buf = new byte[4];

                    //  'RIFF'の確認
                    if (fs.Read(buf, 0, 4) == 4 &&
                        Encoding.GetEncoding(0).GetString(buf) != "RIFF")
                    {
                        return info;
                    }

                    //  ファイルサイズ部分はスキップ
                    fs.Position = fs.Position + 4;

                    //  'WAVE'の確認
                    if (fs.Read(buf, 0, 4) == 4 &&
                        Encoding.GetEncoding(0).GetString(buf) != "WAVE")
                    {
                        return info;
                    }

                    //  各チャンク位置の取得
                    string chunkID = string.Empty;
                    int chunkSize = 0;
                    byte[] chunkIDBuf = new byte[4];
                    byte[] chunkSizeBuf = new byte[4];
                    while (fs.Read(chunkIDBuf, 0, 4) == 4 &&
                           fs.Read(chunkSizeBuf, 0, 4) == 4)
                    {
                        //  チャンク文字・サイズの取得
                        chunkID = Encoding.GetEncoding(0).GetString(chunkIDBuf);
                        chunkSize = BitConverter.ToInt32(chunkSizeBuf, 0);

                        //  各チャンク位置を記憶
                        if (chunkID == "fmt ")
                        {
                            fmtChunkPos = fs.Position;
                        }
                        else if (chunkID == "smpl")
                        {
                            smplChunkPos = fs.Position;
                        }
                        else if (chunkID == "data")
                        {
                            info.SampleChunkPos = fs.Position;
                            info.SampleDataSize = chunkSize;
                        }

                        //  次のチャンクに移動
                        fs.Position = fs.Position + chunkSize;
                    }

                    //  正しいチャンクが取得できているか確認
                    if (fmtChunkPos == 0 || info.SampleChunkPos == 0)
                    {
                        return info;
                    }

                    //  fmtチャンクへ移動
                    fs.Position = fmtChunkPos;

                    //  フォーマットID
                    fs.Read(buf, 0, 2);
                    info.FormatID = BitConverter.ToInt16(buf, 0);

                    //  チャンネル
                    fs.Read(buf, 0, 2);
                    info.ChannelCount = BitConverter.ToInt16(buf, 0);

                    //  レート
                    fs.Read(buf, 0, 4);
                    info.SampleRate = BitConverter.ToInt32(buf, 0);

                    //  1秒あたりのバイト数
                    fs.Read(buf, 0, 4);
                    info.BytesPerSecond = BitConverter.ToInt32(buf, 0);
                    if (info.BytesPerSecond <= 0)
                    {
                        return info;
                    }

                    //  1ブロックあたりのバイト数
                    fs.Read(buf, 0, 2);
                    info.BlockAlign = BitConverter.ToInt16(buf, 0);

                    //  ビット
                    fs.Read(buf, 0, 2);
                    info.Bit = BitConverter.ToInt16(buf, 0);

                    //  時間
                    info.Duration = (double)info.SampleDataSize / info.BytesPerSecond;

                    //  サンプル数
                    info.SampleCount = info.SampleDataSize / (info.ChannelCount * (info.Bit / 8));

                    //  ループ情報の取得
                    if (smplChunkPos != 0)
                    {
                        info.HasLoop = true;

                        fs.Position = smplChunkPos + 44;
                        fs.Read(buf, 0, 4);
                        info.LoopStartSample = BitConverter.ToInt32(buf, 0);
                        fs.Read(buf, 0, 4);
                        info.LoopEndSample = BitConverter.ToInt32(buf, 0) + 1;
                    }

                    //  ここまで取得できたら正しい情報と認定
                    info.IsValid = true;
                }
                catch (Exception e)
                {
                    //  エラーが発生したらメッセージを表示
                    MessageBox.Show(e.Message);
                }
            }

            return info;
        }

        //---------------------------------------------------------------------
        //  波形データの取得（※ PCMモノラル16bit限定）
        //---------------------------------------------------------------------
        public static List<short> GetWaveData(string filename)
        {
            //  フォーマット情報の取得
            var info = GetFormatInfo(filename);

            //  PCMモノラル16ビットのみ対応
            if (!info.IsValid || info.FormatID != 1 ||
                info.Bit != 16 || info.ChannelCount != 1)
            {
                return null;
            }

            //  データリストの作成
            List<short> waveData = new List<short>();

            //  ファイルを読み込んで解析
            using (FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read))
            {
                try
                {
                    const int BUFFER_SIZE = 1024;
                    byte[] buf = new byte[BUFFER_SIZE];
                    int remainSize = info.SampleDataSize;
                    fs.Position = info.SampleChunkPos;

                    //  読み込んだデータを16ビットに変換してデータリストに追加
                    while (remainSize > 0)
                    {
                        int readSize = 0;
                        if (remainSize > BUFFER_SIZE)
                        {
                            readSize = fs.Read(buf, 0, BUFFER_SIZE);
                        }
                        else
                        {
                            readSize = fs.Read(buf, 0, remainSize);
                        }

                        remainSize = remainSize - readSize;

                        for (int i = 0; i < readSize; i = i + 2)
                        {
                            waveData.Add(BitConverter.ToInt16(buf, i));
                        }
                    }
                }
                catch (Exception e)
                {
                    //  エラーが発生したらメッセージを表示
                    MessageBox.Show(e.Message);

                    return null;
                }
            }

            return waveData;
        }

        //---------------------------------------------------------------------
        //  ファイル作成（※ PCMモノラル16bit限定）
        //---------------------------------------------------------------------
        public static bool CreateFile(WaveFormatInfo info, List<short> waveData)
        {
            //  PCMモノラル16ビットのみ対応
            if (!info.IsValid || info.FormatID != 1 ||
                info.Bit != 16 || info.ChannelCount != 1)
            {
                return false;
            }

            using (BinaryWriter bw =
                new BinaryWriter(new FileStream(info.Filename, FileMode.Create)))
            {
                try
                {
                    //  RIFFヘッダ
                    bw.Write(Encoding.GetEncoding(0).GetBytes("RIFF"));
                    bw.Write((int)0);    //  後でファイルサイズ-8を記述する
                    bw.Write(Encoding.GetEncoding(0).GetBytes("WAVE"));

                    //  fmtチャンク
                    bw.Write(Encoding.GetEncoding(0).GetBytes("fmt "));
                    bw.Write((int)16);   //  PCMなので16固定
                    bw.Write(info.FormatID);
                    bw.Write(info.ChannelCount);
                    bw.Write(info.SampleRate);
                    bw.Write(info.BytesPerSecond);
                    bw.Write(info.BlockAlign);
                    bw.Write(info.Bit);

                    //  smplチャンク
                    if (info.HasLoop && info.LoopStartSample >= 0 &&
                        info.LoopEndSample > info.LoopStartSample)
                    {
                        bw.Write(Encoding.GetEncoding(0).GetBytes("smpl"));
                        bw.Write((int)60);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write((int)1);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write((int)0);
                        bw.Write(info.LoopStartSample);
                        bw.Write(info.LoopEndSample - 1);
                        bw.Write((int)0);
                        bw.Write((int)0);
                    }

                    //  dataチャンク
                    bw.Write(Encoding.GetEncoding(0).GetBytes("data"));
                    bw.Write((int)(((info.Bit / 8) * info.ChannelCount) * info.SampleCount));
                    foreach (var data in waveData)
                    {
                        bw.Write(data);
                    }

                    //  ファイルサイズを記述
                    bw.Seek(4, SeekOrigin.Begin);
                    bw.Write((int)(bw.BaseStream.Length - 8));
                }
                catch (Exception e)
                {
                    MessageBox.Show(e.Message);

                    return false;
                }
            }

            return true;
        }
    }
}
